/* micro-lisp */
/* Brian Harvey for CS61c, Berkeley University */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

struct thing {
    char th_type;
    union {
        char *thu_symbol;
        int thu_number;
        struct {struct thing *thup_car, *thup_cdr;} thu_pair;
    } th_union;
};

/* types in th_type */
#define SYMBOL 1
#define NUMBER 2
#define PAIR 3

#define th_symbol th_union.thu_symbol
#define th_number th_union.thu_number
#define th_car th_union.thu_pair.thup_car
#define th_cdr th_union.thu_pair.thup_cdr

typedef struct thing *ThingPtr;

#define NIL (ThingPtr)0

/* constructors */

ThingPtr cons(ThingPtr a, ThingPtr b) {
    ThingPtr new;

    new = (ThingPtr)malloc(sizeof(struct thing));
    new->th_type = PAIR;
    new->th_car = a;
    new->th_cdr = b;
    return(new);
}

ThingPtr make_symbol(char *str) {
    ThingPtr new;

    new = (ThingPtr)malloc(sizeof(struct thing));
    new->th_type = SYMBOL;
    new->th_symbol = str;
    return(new);
}

ThingPtr make_number(int num) {
    ThingPtr new;

    new = (ThingPtr)malloc(sizeof(struct thing));
    new->th_type = NUMBER;
    new->th_number = num;
    return(new);
}


/* selectors */

ThingPtr car(ThingPtr pair) {
    if (pair == NIL || pair->th_type != PAIR) {
        printf("Error: car of non-pair.\n");
        return(NIL);
    }
    return(pair->th_car);
}

ThingPtr cdr(ThingPtr pair) {
    if (pair == NIL || pair->th_type != PAIR) {
        printf("Error: cdr of non-pair.\n");
        return(NIL);
    }
    return(pair->th_cdr);
}

#define cadr(x) (car(cdr(x)))
#define caddr(x) (car(cdr(cdr(x))))

/* printer */

void printthing(ThingPtr th) {
    void printtail(ThingPtr);

    if (th == NIL) printf("()");
    else switch (th->th_type) {
        case SYMBOL:
            printf("%s",th->th_symbol);
	    break;
        case NUMBER:
	    printf("%d",th->th_number);
	    break;
        case PAIR:
	    printf("(");
	    printthing(car(th));
	    printtail(cdr(th));
    }
}

void printtail(ThingPtr th) {
    if (th == NIL) printf(")");
    else switch (th->th_type) {
        case SYMBOL:
	    printf(" . %s)",th->th_symbol);
	    break;
        case NUMBER:
	    printf(" . %d)",th->th_number);
	    break;
        case PAIR:
	    printf(" ");
	    printthing(car(th));
	    printtail(cdr(th));
    }
}


/* reader */

ThingPtr readthing(void) {
    ThingPtr readlist(void), quoted(ThingPtr);
    int IsNumber(char *);
    char ch,buffer[30],*new;

    while (isspace(ch = getchar())) ;
    if (ch == '(') return(readlist());
    if (ch == '\'') return(quoted(readthing()));
    if (ch == ')') {
	fprintf(stderr,"Unmatched close paren.\n");
	return(NIL);
    }
    ungetc(ch, stdin);
    scanf("%29[^() \n\t]",buffer);
    if (IsNumber(buffer)) return(make_number(atoi(buffer)));
    new = (char *)malloc(1+strlen(buffer));
    strcpy(new,buffer);
    return(make_symbol(new));
}

int IsNumber(char *str) {
    char ch;

    while ((ch = *str++)) {
        if (ch != '-' && (ch < '0' || ch > '9'))
            return(0);
    }
    return(1);
}

ThingPtr quoted(ThingPtr th) {
    return(cons(make_symbol("quote"),cons(th,NIL)));
}

ThingPtr readlist(void) {
    ThingPtr new;
    char ch;

    while (isspace(ch = getchar())) ;
    if (ch == ')') return(NIL);
    ungetc(ch,stdin);
    new = readthing();
    if (new != NIL && new->th_type == SYMBOL && !strcmp(new->th_symbol,".")) {
        new = readthing();
        ch = getchar();
        if (ch != ')')
            printf("Error in dot notation.\n");
        return(new);
    }
    return(cons(new,readlist()));
}


/* vast collection of primitives */

ThingPtr plus(ThingPtr a, ThingPtr b) {
    if (a == NIL || a->th_type != NUMBER
                 || b == NIL || b->th_type != NUMBER) {
        printf("Non-numeric arg to plus.\n");
        return(NIL);
    }
    return(make_number(a->th_number + b->th_number));
}

/* eval */

ThingPtr eval(ThingPtr exp) {
    char *fn;

    if (exp == NIL) return(NIL);
    switch(exp->th_type) {
        case NUMBER:
	    return(exp);
        case SYMBOL:
	    printf("What do you expect, variables???\n");
	    return(NIL);
        case PAIR:
	    if ((car(exp))->th_type != SYMBOL) {
		printf("Application of non-procedure\n");
		return(NIL);
	    }
	    fn = (car(exp))->th_symbol;
	    if (!strcmp(fn,"quote"))
		return(cadr(exp));
	    if (!strcmp(fn,"+"))
		return(plus(eval(cadr(exp)),eval(caddr(exp))));
	    if (!strcmp(fn,"car"))
		return(car(eval(cadr(exp))));
	    if (!strcmp(fn,"cdr"))
		return(cdr(eval(cadr(exp))));
	    if (!strcmp(fn,"cons"))
		return(cons(eval(cadr(exp)),eval(caddr(exp))));
	    if (!strcmp(fn,"exit"))
		exit(0);
	    printf("Unknown function\n");
	    return(NIL);
    }
}

/* read-eval-print loop */

int main() {
    for (;;) {
        printf("> ");
        printthing(eval(readthing()));
        printf("\n");
    }
    return 0;
}
